syntax = "proto3";

package test;

service Test {
  rpc UnaryCall(Input) returns (Output);
}

message Input {}
message Output {}

service Test1 {
  rpc UnaryCall(Input1) returns (Output1);

  rpc StreamCall(Input1) returns (stream Output1);
}

message Input1 {
  bytes buf = 1;
}
message Output1 {
  bytes buf = 1;
}
