syntax = "proto3";

package test;

service Test {
  rpc TestRequest(SomeData) returns (SomeData);
}

message SomeData {
  // include a bunch of data so there actually is something to compress
  bytes data = 1;
}
